import importlib.resources as res
import operator
import re
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from functools import reduce
from typing import Any, Iterator, TypeAlias, TypeGuard, TypeVar, overload

import elasticai.creator.function_utils as F
import elasticai.creator.plugin as _pl
from elasticai.creator.function_utils import KeyedFunctionDispatcher
from elasticai.creator.graph import BaseGraph
from elasticai.creator.ir import (
    Attribute,
    LoweringPass,
    RequiredField,
)
from elasticai.creator.ir import Edge as _Edge
from elasticai.creator.ir import (
    Implementation as _Implementation,
)
from elasticai.creator.ir import Node as _Node
from elasticai.creator.plugin import PluginLoader as _Loader
from elasticai.creator.plugin import PluginSpec as _PluginSpec
from elasticai.creator.plugin import PluginSymbol as _PluginSymbol


@dataclass
class PluginSpec(_PluginSpec):
    generated: tuple[str, ...]
    static_files: tuple[str, ...]


ShapeTuple: TypeAlias = tuple[int] | tuple[int, int] | tuple[int, int, int]


def is_shape_tuple(values) -> TypeGuard[ShapeTuple]:
    max_num_values = 3
    return len(values) <= max_num_values


class Shape:
    @overload
    def __init__(self, width: int, /) -> None: ...

    @overload
    def __init__(self, depth: int, width: int, /) -> None: ...

    @overload
    def __init__(self, depth: int, width: int, height: int, /) -> None: ...

    def __init__(self, *values: int) -> None:
        """values are interpreted as one of the following:
        - width
        - depth, width
        - depth, width, height

        Usually width is kernel_size, depth is channels
        """

        if is_shape_tuple(values):
            self._data = values
        else:
            raise TypeError(f"taking at most three ints, given {values}")

    @classmethod
    def from_tuple(cls, values: ShapeTuple | list[int]) -> "Shape":
        return cls(*values)  # type ignore

    def to_tuple(self) -> ShapeTuple:
        return self._data

    def to_list(self) -> list[int]:
        return list(self.to_tuple())

    def __getitem__(self, item):
        return self._data[item]

    def size(self) -> int:
        return reduce(operator.mul, self._data, 1)

    def ndim(self) -> int:
        return len(self._data)

    @property
    def depth(self) -> int:
        if len(self._data) > 1:
            return self._data[0]
        return 1

    def __eq__(self, other):
        if isinstance(other, tuple):
            return self._data == other
        if isinstance(other, Shape):
            return self._data == other._data

        return False

    @property
    def width(self) -> int:
        if len(self._data) > 1:
            return self._data[1]
        else:
            return self._data[0]

    @property
    def height(self) -> int:
        if len(self._data) > 2:
            return self._data[2]
        return 1

    def __repr__(self) -> str:
        match self._data:
            case (width,):
                return f"Shape({width=})"
            case (depth, width):
                return f"Shape({depth=}, {width=})"
            case (depth, width, height):
                return f"Shape({depth=}, {width=}, {height=})"
            case _:
                return f"Shape({self._data})"


class ShapeField(RequiredField[list[int], Shape]):
    def __init__(self):
        super().__init__(
            set_convert=lambda x: x.to_list(), get_convert=Shape.from_tuple
        )


class VhdlNode(_Node):
    """Extending ir.core.Node to a vhdl specific node.

    `VhdlNode` contains all knowledge that we need to create
    and use an instance of a vhdl entity. However, this becomes
    a little bit complicated because vhdl differentiates between
    the *entity* and the *architecture* of a component.
    The entity is similar to an _interface_ while the architecture
    is similar to the _implementation_. However, to instantiate
    components, we need to know both names.

    Attributes:

    implementation:: The name of the implementation will be used to derive the architecture name.
        E.g., if the implementation is `"adder"`, we will instantiate the entity `work.adder(rtl)`.
        CAUTION: This behaviour is subject to change. Future versions might require the full entity name
    """

    implementation: str
    input_shape: RequiredField[list[int], Shape] = ShapeField()
    output_shape: RequiredField[list[int], Shape] = ShapeField()


def vhdl_node(
    name: str,
    type: str,
    implementation: str,
    input_shape: Shape | ShapeTuple,
    output_shape: Shape | ShapeTuple,
    attributes: dict | None = None,
) -> VhdlNode:
    """Convenience method for creating new vhdl nodes."""
    if attributes is None:
        attributes = {}

    def to_tuple(s: Shape | ShapeTuple) -> ShapeTuple:
        if isinstance(s, Shape):
            return s.to_tuple()
        else:
            return s

    return VhdlNode(
        name=name,
        data=dict(
            type=type,
            implementation=implementation,
            input_shape=to_tuple(input_shape),
            output_shape=to_tuple(output_shape),
        )
        | attributes,
    )


class Edge(_Edge):
    src_dst_indices: tuple[tuple[int, int] | tuple[str, str], ...]

    def __hash__(self) -> int:
        return hash((self.src, self.dst, self.src_dst_indices))


def edge(
    src: str, dst: str, src_dst_indices: Iterable[tuple[int, int]] | tuple[str, str]
) -> Edge:
    return Edge(src=src, dst=dst, data={"src_dst_indices": tuple(src_dst_indices)})


N = TypeVar("N", bound=VhdlNode)
E = TypeVar("E", bound=Edge)


class Implementation(_Implementation[N, E]):
    name: str
    type: str

    @overload
    def __init__(
        self: "Implementation[VhdlNode, Edge]",
        *,
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
        attributes: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[N, Edge]",
        *,
        node_fn: Callable[[dict], N],
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
        attributes: dict[str, Attribute] | None = None,
        graph: BaseGraph | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[VhdlNode, E]",
        *,
        edge_fn: Callable[[dict], E],
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
        attributes: dict[str, Attribute] | None = None,
        graph: BaseGraph | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[N, E]",
        *,
        node_fn: Callable[[dict], N],
        edge_fn: Callable[[dict], E],
        name: str | None = None,
        type: str | None = None,
        data: dict[str, Attribute] | None = None,
        attributes: dict[str, Attribute] | None = None,
        graph: BaseGraph | None = None,
    ) -> None: ...

    def __init__(
        self,
        *,
        name: str | None = None,
        type: str | None = None,
        node_fn=VhdlNode,
        edge_fn=Edge,
        data: dict[str, Any] | None = None,
        attributes: dict[str, Any] | None = None,
        graph: BaseGraph | None = None,
    ) -> None:
        if attributes is not None and data is not None:
            raise TypeError("pass either attributes or data argument")
        if attributes is not None:
            warnings.warn(
                "the argument `attributes` is deprecated, use `data` instead.",
                DeprecationWarning,
                2,
            )
            data = attributes

        if graph is None:
            graph = BaseGraph()
        super().__init__(
            node_fn=node_fn,
            edge_fn=edge_fn,
            data=data,
            graph=graph,
        )
        if name is not None:
            self.data["name"] = name
        if type is not None:
            self.data["type"] = type


Code: TypeAlias = tuple[str, Sequence[str]]


class Ir2Vhdl(LoweringPass[Implementation, Code]):
    def __init__(self) -> None:
        super().__init__()
        self.__static_files: dict[str, Callable[[], str]] = {}

    def register_static(self, name: str, fn: Callable[[], str]) -> None:
        self.__static_files[name] = fn

    def __call__(self, args: Iterable[Implementation]) -> Iterator[Code]:
        for name, content in super().__call__(args):
            yield f"{name}.vhd", content
        for name, fn in self.__static_files.items():
            yield name, [fn()]


class Signal(ABC):
    types: set[type["Signal"]] = set()

    @abstractmethod
    def define(self) -> Iterator[str]: ...

    @property
    @abstractmethod
    def name(self) -> str: ...

    @classmethod
    def from_code(cls, code: str) -> "Signal":
        for t in cls.types:
            if t.can_create_from_code(code):
                return t.from_code(code)
        return NullDefinedLogicSignal.from_code(code)

    @classmethod
    @abstractmethod
    def can_create_from_code(cls, code: str) -> bool: ...

    @classmethod
    def register_type(cls, t: type["Signal"]) -> None:
        cls.types.add(t)

    @abstractmethod
    def make_instance_specific(self, instance: str) -> "Signal": ...


class LogicSignal(Signal):
    def __init__(self, name: str):
        self._name = name

    def define(self) -> Iterator[str]:
        yield f"signal {self._name} : std_logic := '0';"

    @property
    def name(self) -> str:
        return self._name

    @classmethod
    def can_create_from_code(cls, code: str) -> bool:
        return cls._search(code) is not None

    @classmethod
    def _search(cls, code: str) -> re.Match[str] | None:
        match = re.search(
            r"signal ([a-zA-Z_][a-zA-Z0-9_]*)\s*:\s*std_logic(?:\s+|;)", code
        )
        return match

    @classmethod
    def from_code(cls, code: str) -> "Signal":
        match = cls._search(code)
        if match is None:
            raise ValueError(f"Cannot create signal from code: {code}")
        (name,) = match.groups()
        return cls(name)

    def make_instance_specific(self, instance: str) -> Signal:
        return self.__class__(f"{self.name}_{instance}")

    def __eq__(self, other: object) -> bool:
        if other is self:
            return True
        if isinstance(other, LogicSignal):
            return self._name == other._name
        return False


class LogicVectorSignal(Signal):
    def __init__(self, name: str, width: int):
        self._name = name
        self._width = width

    def define(self) -> Iterator[str]:
        yield f"signal {self._name} : std_logic_vector({self._width} - 1 downto 0) := (others => '0');"

    @property
    def name(self) -> str:
        return self._name

    @property
    def width(self) -> int:
        return self._width

    @classmethod
    def can_create_from_code(cls, code: str) -> bool:
        return cls._search(code) is not None

    @classmethod
    def _search(cls, code: str) -> re.Match[str] | None:
        match = re.match(
            r"signal ([a-zA-Z_][a-zA-Z0-9_]*)\s*: std_logic_vector\((\d+|(?:\d+ - \d+)) downto 0\)",
            code,
        )
        return match

    @classmethod
    def from_code(cls, code: str) -> "Signal":
        match = cls._search(code)
        if match is None:
            raise ValueError(f"Cannot create signal from code: {code}")
        name, width = match.groups()
        if " - " in width:
            a, b = width.split(" - ")
            width = str(int(a) - int(b))
        return cls(name, int(width) + 1)

    def make_instance_specific(self, instance: str) -> Signal:
        return self.__class__(f"{self.name}_{instance}", self.width)

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        if isinstance(other, LogicVectorSignal):
            return self._name == other._name and self._width == other._width
        return False


class NullDefinedLogicSignal(Signal):
    def __init__(self, name):
        self._name = name

    def define(self) -> Iterator[str]:
        yield from []

    @property
    def name(self) -> str:
        return self._name

    @classmethod
    def can_create_from_code(cls, code: str) -> bool:
        return False

    @classmethod
    def from_code(cls, code: str) -> "Signal":
        return cls("<unknown>")

    def make_instance_specific(self, instance: str) -> Signal:
        return self


for t in (LogicSignal, LogicVectorSignal, NullDefinedLogicSignal):
    Signal.register_type(t)


class PortMap:
    def __init__(self, map: dict[str, Signal]):
        self._signals: dict[str, Signal] = map

    def as_dict(self) -> dict[str, str]:
        return {k: tuple(v.define())[0] for k, v in self._signals.items()}

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> "PortMap":
        return cls({k: Signal.from_code(v) for k, v in data.items()})

    def __eq__(self, other: object) -> bool:
        if other is self:
            return True
        if isinstance(other, PortMap):
            return self._signals == other._signals
        return False


class Instance:
    """Represents an entity that we can/want instantiate.

    The aggregates all the knowledge that is necessary to
    instantiate and use the corresponding entity programmatically,
    when generating vhdl code.
    """

    def __init__(
        self,
        node: VhdlNode,
        generic_map: dict[str, str],
        port_map: dict[str, Signal],
    ):
        self._node = node
        self._generics: dict[str, str] = {k.lower(): v for k, v in generic_map.items()}
        self.port_map = {
            k: v.make_instance_specific(self._node.name) for k, v in port_map.items()
        }

    @property
    def input_shape(self) -> Shape:
        return self._node.input_shape

    @property
    def output_shape(self) -> Shape:
        return self._node.output_shape

    @property
    def name(self) -> str:
        return self._node.name

    @property
    def implementation(self) -> str:
        return self._node.implementation

    def define_signals(self) -> Iterator[str]:
        for s in self.port_map.values():
            yield from s.define()

    def instantiate(self) -> Iterator[str]:
        yield from (f"{self.name}: entity work.{self.implementation}(rtl) ",)
        generics = tuple(self._generics.items())
        if len(generics) > 0:
            yield "generic map ("
            for key, value in generics[:-1]:
                yield f"  {key.upper()} => {value},"
            for g in generics[-1:]:
                yield f"  {g[0].upper()} => {g[1]}"
            yield "  )"
        port_map = tuple(self.port_map.items())
        yield "  port map ("

        for k, v in port_map[:-1]:
            yield f"    {k} => {v.name},"
        for k, v in port_map[-1:]:
            yield f"    {k} => {v.name}"

        yield "  );"


class InstanceFactory(KeyedFunctionDispatcher[VhdlNode, Instance]):
    """Automatically creates Instances from VhdlNodes based on their `type` field."""

    def __init__(self):
        def dispatch_key_fn(node: VhdlNode) -> str:
            return node.type

        super().__init__(dispatch_key_fn=dispatch_key_fn)


PluginSymbol: TypeAlias = _PluginSymbol[Ir2Vhdl]


class PluginLoader(_Loader[Ir2Vhdl]):
    """Plugin loader for ir2vhdl translation."""

    def __init__(self, lowering: Ir2Vhdl):
        builder: _pl.SymbolFetcherBuilder[PluginSpec, Ir2Vhdl] = (
            _pl.SymbolFetcherBuilder(PluginSpec)
        )
        fetcher: _pl.SymbolFetcher[Ir2Vhdl] = (
            builder.add_fn(self.__get_generated)
            .add_fn(_StaticFile.make_symbols)
            .build()
        )
        super().__init__(
            fetch=fetcher,
            plugin_receiver=lowering,
        )

    def load_from_package(self, package: str) -> None:
        if "." not in package:
            package = f"elasticai.creator_plugins.{package}"
        super().load_from_package(package)

    @staticmethod
    def __get_generated(plugin: PluginSpec) -> Iterator[PluginSymbol]:
        if plugin.target_runtime == "vhdl":
            yield from _pl.import_symbols(plugin.package, plugin.generated)


class _StaticFile(_PluginSymbol[Ir2Vhdl]):
    def __init__(self, name: str, package: str):
        self._name = name
        self._package = package

    @property
    def name(self) -> str:
        return self._name

    def load_into(self, receiver: Ir2Vhdl):
        receiver.register_static(self.name, self)

    @classmethod
    def make_symbols(cls, p: PluginSpec) -> Iterator[PluginSymbol]:
        if p.target_runtime == "vhdl":
            for name in p.static_files:
                yield cls(name=name, package=p.package)

    def __call__(self) -> str:
        file = res.files(self._package).joinpath(f"vhdl/{self.name}")
        return file.read_text()


_Tcontra = TypeVar("_Tcontra", contravariant=True)

TypeHandlerFn: TypeAlias = Callable[[Implementation], Code]


def _type_handler(name: str, fn: TypeHandlerFn) -> PluginSymbol:
    def load_into(lower: Ir2Vhdl) -> None:
        lower.register(name)(fn)

    return _pl.make_plugin_symbol(load_into, fn)


def _type_handler_for_iterable(
    name: str, fn: Callable[[Implementation], Iterable[Code]]
) -> PluginSymbol:
    def load_into(lower: Ir2Vhdl) -> None:
        lower.register_iterable(name)(fn)

    return _pl.make_plugin_symbol(load_into, fn)


type_handler = F.FunctionDecorator(_type_handler)
type_handler_iterable = F.FunctionDecorator(_type_handler_for_iterable)



from elasticai.creator.ir2vhdl import Implementation, edge, vhdl_node
from elasticai.creator_plugins.grouped_filter import FilterParameters


def network(input_shape, kernel_size, out_channels, stride=2):
    out_length = (input_shape[1] - kernel_size) // stride + 1
    in_channels = input_shape[0]
    _net = Implementation(name="network", type="network")
    _net.add_nodes(
        (
            vhdl_node(
                name="input",
                input_shape=input_shape,
                output_shape=input_shape,
                type="input",
                implementation="",
            ),
            vhdl_node(
                name="conv0_i0",
                type="filter",
                implementation="conv0",
                input_shape=input_shape,
                output_shape=(out_channels, out_length),
                attributes=dict(
                    filter_parameters=FilterParameters(
                        kernel_size=kernel_size,
                        in_channels=in_channels,
                        out_channels=out_channels,
                        stride=stride,
                    ).as_dict(),
                ),
            ),
            vhdl_node(
                name="conv1_i0",
                type="filter",
                implementation="conv1",
                input_shape=(out_channels, out_length),
                output_shape=(out_channels, out_length - kernel_size + 1),
                attributes=dict(
                    filter_parameters=FilterParameters(
                        kernel_size=kernel_size,
                        in_channels=out_channels,
                        out_channels=out_channels,
                        stride=1,
                    ).as_dict()
                ),
            ),
            vhdl_node(
                name="output",
                type="output",
                implementation="",
                input_shape=(out_channels, out_length - kernel_size + 1),
                output_shape=(out_channels, out_length - kernel_size + 1),
            ),
        )
    )
    _net.add_edges(
        (
            edge(src="input", dst="conv0_i0", src_dst_indices=tuple()),
            edge(src="conv0_i0", dst="conv1_i0", src_dst_indices=tuple()),
            edge(src="conv1_i0", dst="output", src_dst_indices=tuple()),
        )
    )

    return _net



from dataclasses import dataclass
from typing import Iterator


@dataclass
class Port:
    inputs: dict[str, str]
    outputs: dict[str, str]

    def signals(self) -> Iterator[str]:
        for s, type in self.inputs.items():
            yield f"signal {s} : in {type}"
        for s, type in self.outputs.items():
            yield f"signal {s} : out {type}"


class VHDLEntity:
    def __init__(self, name: str, port: Port, generics: dict[str, str]):
        self._name = name
        self._generics = generics
        self._port = port

    def _generate_generic(self):
        if len(self._generics) > 0:
            yield "generic ("
            generics = tuple(self._generics.items())
            for generic, type in generics[:-1]:
                yield f"{generic} : {type};"
            generic, type = generics[-1]
            yield f"{generic} : {type}"
            yield ");"

    def _generate_port(self):
        yield "port ("
        signals = tuple(self._port.signals())
        for signal in signals[:-1]:
            yield f"{signal};"
        yield signals[-1]
        yield ");"

    def _generate_library_clause(self):
        yield from ("library ieee;", "use ieee.std_logic_1164.all;")

    def generate_entity(self) -> Iterator[str]:
        yield from self._generate_library_clause()
        yield f"entity {self._name} is"
        yield from self._generate_generic()
        yield from self._generate_port()
        yield "end entity;"



from elasticai.creator.ir2vhdl import VhdlNode

from .node_factory import (
    InstanceFactoryForCombinatorial,
)
from .shift_register import ShiftRegister


@InstanceFactoryForCombinatorial.register
def striding_shift_register(node: VhdlNode):
    return StridingShiftRegister(node)


class StridingShiftRegister(ShiftRegister):
    def __init__(
        self,
        node: VhdlNode,
    ):
        super().__init__(node)
        if (
            "generic_map" in node.attributes  # pyright: ignore
            and "stride" in node.attributes["generic_map"]  # pyright: ignore
        ):
            stride = node.attributes["generic_map"]["stride"]  # pyright: ignore
        else:
            stride = node.attributes["stride"]  # pyright: ignore
        self._generics["stride"] = stride



from collections.abc import Mapping
from typing import Any

from .rewriter import (
    CompositeRule,
    DataGraph,
    IrFactory,
    Pattern,
    PatternRuleSpec,
    ReadOnlyDataGraph,
    Rule,
    StdPattern,
)
from .rewriter import (
    PatternRule as _PatternRule,
)

__all__ = [
    "Rewriter",
    "PatternRuleSpec",
    "ReadOnlyDataGraph",
    "IrFactory",
    "Rule",
    "DataGraph",
    "CompositeRule",
    "PatternRule",
    "Pattern",
    "StdPattern",
]

from elasticai.creator.graph import BaseGraph as _BaseGraph

from ..core import Edge as _Edge
from ..core import Implementation as _Implementation
from ..core import Node as _Node


class _DefaultFactory(IrFactory):
    def node(self, name: str, data: Mapping[str, Any]) -> _Node:
        return _Node(name=name, data=dict(data))

    def edge(self, src: str, dst: str, data: Mapping[str, Any]) -> _Edge:
        return _Edge(src=src, dst=dst, data=dict(data))

    def data_graph(self) -> DataGraph[_Node, _Edge]:
        return _Implementation(graph=_BaseGraph(), data={})


class PatternRule(_PatternRule):
    def __init__(self, spec: PatternRuleSpec):
        super().__init__(ir_factory=_DefaultFactory(), spec=spec)



from collections.abc import Sequence
from itertools import chain

from elasticai.creator import graph as gr
from elasticai.creator.ir2vhdl import Implementation, type_handler

from .combinatorial import (
    build_data_signal_connections_for_combinatorial,
    build_declarations_for_combinatorial,
    build_instantiations_for_combinatorial,
    wrap_in_architecture,
)
from .language import Port, VHDLEntity


@type_handler
def clocked_combinatorial(impl: Implementation) -> tuple[str, Sequence[str]]:
    def _iter():
        input_size = impl.nodes["input"].input_shape.size()
        output_size = impl.nodes["output"].output_shape.size()
        entity = VHDLEntity(
            name=impl.name,
            port=Port(
                inputs=dict(
                    clk="std_logic",
                    rst="std_logic",
                    d_in=f"std_logic_vector({input_size} - 1 downto 0)",
                    valid_in="std_logic",
                ),
                outputs=dict(
                    d_out=f"std_logic_vector({output_size} - 1 downto 0)",
                    valid_out="std_logic",
                ),
            ),
            generics=dict(),
        )
        yield from entity.generate_entity()
        yield ""
        declarations = build_declarations_for_combinatorial(impl)
        declarations = chain(
            declarations,
            (
                "signal valid_out_input : std_logic := '0';",
                "signal valid_in_output : std_logic := '0';",
            ),
        )
        definitions = build_data_signal_connections_for_combinatorial(impl)
        connected_valid_signals = []
        valid_in_out_pairs = tuple(_get_valid_in_out_pairs(impl).items())
        connected_valid_signals.extend(("valid_out_input <= valid_in;",))
        for dst, src in valid_in_out_pairs:
            connected_valid_signals.append(f"valid_in_{dst} <= valid_out_{src};")
        connected_valid_signals.append("valid_out <= valid_in_output;")
        definitions = chain(
            definitions,
            connected_valid_signals,
            ("", ""),
            build_instantiations_for_combinatorial(impl),
        )

        definitions = chain(
            definitions,
        )
        yield from wrap_in_architecture(impl.name, declarations, definitions)

    return impl.name, tuple(_iter())


def _get_valid_in_out_pairs(impl: Implementation) -> dict[str, str]:
    def is_clocked(node):
        return node.type in [
            "striding_shift_register",
            "sliding_window",
            "shift_register",
        ]

    def iterate(node: str):
        def pred(node: str):
            return impl.predecessors(node)

        def succ(node: str):
            return impl.successors(node)

        for node in gr.bfs_iter_up(successors=succ, predecessors=pred, start=node):
            yield impl.nodes[node]

    adjacency: dict[str, str] = {}
    last_clocked_node = "input"
    for node in filter(is_clocked, impl.nodes.values()):
        last_clocked_node = node.name
        if len(adjacency) == 0:
            adjacency[node.name] = "input"
        else:
            for pred in filter(
                is_clocked,
                iterate(node.name),
            ):
                adjacency[node.name] = pred.name
                break
    adjacency["output"] = last_clocked_node
    return adjacency



from os.path import exists
from pathlib import Path

import torch

from elasticai.creator.arithmetic import (
    FxpArithmetic,
)
from elasticai.creator.file_generation import find_project_root
from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.testing import (
    build_report_folder_and_testdata,
    run_cocotb_sim,
)


def routine_testing_precomputed_module(
    dut,
    num_steps: int,
    fxp: FxpArithmetic,
    file_name: str,
    file_suffix: str = "vhd",
) -> None:
    val_input = fxp.as_rational(
        fxp.round_to_integer(
            torch.linspace(
                start=fxp.minimum_as_rational,
                end=fxp.maximum_as_rational,
                steps=2 * num_steps,
            )
        )
    )
    dut.eval()
    with torch.no_grad():
        val_output = dut(val_input)

    output_dir = build_report_folder_and_testdata(
        dut_name=file_name,
        testdata={
            "in": fxp.cut_as_integer(val_input).int().tolist(),
            "out": fxp.round_to_integer(val_output).int().tolist(),
        },
    )

    output_dir_sub = Path(*output_dir.parts[len(find_project_root().parts) :])
    destination = OnDiskPath(str(output_dir_sub), parent=find_project_root())
    dut.create_design(file_name).save_to(destination)
    assert exists(output_dir / f"{file_name}.{file_suffix}")
    assert exists(output_dir / "testdata.json")

    run_cocotb_sim(
        src_files=[output_dir / f"{file_name}.{file_suffix}"],
        top_module_name=file_name,
        cocotb_test_module="tests.integration_tests.nn.fixed_point.precomputed_tb",
        waveform_save_dst=str(output_dir),
    )



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)


class ENV5ReconfigTop:
    def save_to(self, destination: Path):
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="env5_reconfig_top.vhd",
            parameters={},
        )
        file = destination.as_file(".vhd")
        file.write(template)



import copy

from elasticai.creator.graph import BaseGraph
from elasticai.creator.ir import (
    Implementation,
    Node,
    PatternRuleSpec,
    edge,
    node,
)
from elasticai.creator.ir.rewriting import (
    DataGraph,
    Pattern,
    PatternRule,
    ReadOnlyDataGraph,
    StdPattern,
)


class DummyPattern(Pattern):
    def __init__(self, graph, interface, match):
        self._graph = graph
        self._interface = interface
        self._match = match

    @property
    def graph(self) -> DataGraph:
        return self._graph

    @property
    def interface(self) -> set[str]:
        return self._interface

    def match(self, g: ReadOnlyDataGraph) -> list[dict[str, str]]:
        return self._match


def test_raise_error_when_interface_is_not_in_replacement():
    pattern_graph = make_pattern().add_node(node("more_interface", "interface"))
    network = (
        make_network()
        .add_node(node("some_node", "some_type"))
        .add_edges((edge("input", "some_node"), edge("some_node", "output")))
    )

    def replacement(_: ReadOnlyDataGraph):
        return make_replacement().add_edge(edge("start", "end"))

    pattern = DummyPattern(
        graph=pattern_graph,
        interface={"start", "end", "more_interface"},
        match=[{"start": "input", "end": "output", "more_interface": "some_node"}],
    )

    rule = PatternRule(
        PatternRuleSpec(
            pattern=pattern,
            replacement_fn=replacement,
        )
    )

    try:
        rule.apply(network)
    except ValueError as e:
        assert str(e) == "Replacement is missing interface nodes: {'more_interface'}"
    else:
        assert False, "Expected ValueError was not raised."


def test_raise_error_when_interface_is_not_in_pattern():
    pattern_graph = make_pattern()
    network = (
        make_network()
        .add_node(node("some_node", "some_type"))
        .add_edges((edge("input", "some_node"), edge("some_node", "output")))
    )

    def replacement(_: ReadOnlyDataGraph):
        return (
            make_replacement()
            .add_node(node("more_interface", "interface"))
            .add_edge(edge("start", "more_interface"))
            .add_edge(edge("more_interface", "end"))
        )

    pattern = DummyPattern(
        graph=pattern_graph,
        interface={"start", "end", "more_interface"},
        match=[{"start": "input", "end": "output"}],
    )

    rule = PatternRule(
        PatternRuleSpec(
            pattern=pattern,
            replacement_fn=replacement,
        )
    )

    try:
        rule.apply(network)
    except ValueError as e:
        assert str(e) == "Pattern Graph is missing interface nodes: {'more_interface'}"
    else:
        assert False, "Expected ValueError was not raised."


def test_replace_prelu():
    impl = (
        make_network()
        .add_node(node("activation0", "prelu"))
        .add_edge(edge("input", "activation0"))
        .add_edge(edge("activation0", "output"))
    )
    pattern_graph = (
        make_pattern()
        .add_node(node("prelu", "prelu"))
        .add_edge(edge("start", "prelu"))
        .add_edge(edge("prelu", "end"))
    )

    def replacement(match):
        return (
            make_replacement()
            .add_node(
                node(
                    "binarize",
                    "binarize",
                    {
                        "implementation": "binarize",
                    },
                )
            )
            .add_edge(edge("start", "binarize"))
            .add_edge(edge("binarize", "end"))
        )

    def node_constraint(pattern_node: Node, original_node: Node) -> bool:
        if pattern_node.name in ("start", "end"):
            return True
        return pattern_node.type == original_node.type

    pattern = StdPattern(
        graph=pattern_graph,
        node_constraint=node_constraint,
        interface={"start", "end"},
    )

    rule = PatternRule(
        PatternRuleSpec(
            pattern=pattern,
            replacement_fn=replacement,
        )
    )
    new_impl = rule.apply(impl)
    assert len(new_impl.graph.nodes) == 3
    assert "binarize" in new_impl.nodes
    assert new_impl.nodes["binarize"].type == "binarize"


def test_merge_layers():
    impl = (
        make_network()
        .add_nodes(
            (
                node("conv0", "conv1d", {"parameters": {"weight": 4}}),
                node("bnorm0", "batchnorm1d", {"parameters": {"weight": 5}}),
                node("activation0", "relu"),
            )
        )
        .add_edges(
            (
                edge("input", "conv0"),
                edge("conv0", "bnorm0"),
                edge("bnorm0", "activation0"),
                edge("activation0", "output"),
            )
        )
    )
    pattern = (
        make_pattern()
        .add_nodes(
            (node("conv", "conv1d"), node("bnorm", "batchnorm1d"), node("act", "relu"))
        )
        .add_edges(
            (
                edge("start", "conv"),
                edge("conv", "bnorm"),
                edge("bnorm", "act"),
                edge("act", "end"),
            )
        )
    )

    def replacement(match):
        def fuse_conv():
            bnorm = match.nodes["bnorm"].data
            conv1d = match.nodes["conv"].data
            fused = copy.deepcopy(conv1d)

            def dummy_combine_bnorm_and_conv_weights(conv_weights, bnorm_weights):
                return conv_weights + bnorm_weights

            fused["parameters"]["weight"] = dummy_combine_bnorm_and_conv_weights(
                conv1d["parameters"]["weight"], bnorm["parameters"]["weight"]
            )
            return fused

        return (
            make_replacement()
            .add_node(Node("fused_conv", fuse_conv()))
            .add_edge(edge("start", "fused_conv"))
            .add_edge(edge("fused_conv", "end"))
        )

    def node_constraint(pattern_node: Node, original_node: Node) -> bool:
        if pattern_node.name in ("start", "end"):
            return True
        constraint = pattern_node.type == original_node.type
        return constraint

    rule = PatternRule(
        PatternRuleSpec(
            pattern=StdPattern(
                pattern, node_constraint=node_constraint, interface={"start", "end"}
            ),
            replacement_fn=replacement,
        )
    )
    new_impl = rule.apply(impl)
    assert "fused_conv" in set(new_impl.nodes.keys())
    assert "conv0" not in set(new_impl.nodes.keys())


def make_network() -> Implementation:
    return (
        Implementation(graph=BaseGraph(), data={"name": "root", "type": "network"})
        .add_node(node("input", "input"))
        .add_node(node("output", "output"))
    )


def make_pattern() -> Implementation:
    return (
        Implementation(graph=BaseGraph(), data={"name": "pattern", "type": "pattern"})
        .add_node(node("start", "interface"))
        .add_node(node("end", "interface"))
    )


def make_replacement() -> Implementation:
    return (
        Implementation(
            graph=BaseGraph(), data={"name": "replacement", "type": "replacement"}
        )
        .add_node(node("start", "interface"))
        .add_node(node("end", "interface"))
    )


def test_remove_layers_from_matches_with_overlapping_interface_nodes():
    impl = (
        make_network()
        .add_nodes(
            (
                node("conv0", "conv1d"),
                node("bnorm0", "batchnorm1d"),
                node("activation0", "relu"),
                node("conv1", "conv1d"),
                node("bnorm1", "batchnorm1d"),
                node("activation1", "relu"),
            )
        )
        .add_edges(
            (
                edge("input", "conv0"),
                edge("conv0", "bnorm0"),
                edge("bnorm0", "activation0"),
                edge("activation0", "conv1"),
                edge("conv1", "bnorm1"),
                edge("bnorm1", "activation1"),
                edge("activation1", "output"),
            )
        )
    )
    pattern = (
        make_pattern()
        .add_nodes(
            (node("conv", "conv1d"), node("bnorm", "batchnorm1d"), node("act", "relu"))
        )
        .add_edges(
            (
                edge("start", "conv"),
                edge("conv", "bnorm"),
                edge("bnorm", "act"),
                edge("act", "end"),
            )
        )
    )

    def replacement(match):
        return (
            make_replacement()
            .add_node(node("fused_conv", "conv1d"))
            .add_node(node("activation", "relu"))
            .add_edge(edge("start", "fused_conv"))
            .add_edge(edge("fused_conv", "activation"))
            .add_edge(edge("activation", "end"))
        )

    def node_constraint(pattern_node: Node, original_node: Node) -> bool:
        if pattern_node.name in ("start", "end"):
            return True
        constraint = pattern_node.type == original_node.type
        return constraint

    spec = PatternRuleSpec(
        pattern=StdPattern(
            pattern, node_constraint=node_constraint, interface={"start", "end"}
        ),
        replacement_fn=replacement,
    )
    rule = PatternRule(spec)

    new_impl = rule.apply(impl)
    assert len(tuple(new_impl.nodes)) == len(tuple(impl.nodes)) - 1


def test_remove_layers_from_matches_with_overlapping_interface_nodes_with_injected_match():
    impl = (
        make_network()
        .add_nodes(
            (
                node("conv0", "conv1d"),
                node("bnorm0", "batchnorm1d"),
                node("activation0", "relu"),
                node("conv1", "conv1d"),
                node("bnorm1", "batchnorm1d"),
                node("activation1", "relu"),
            )
        )
        .add_edges(
            (
                edge("input", "conv0"),
                edge("conv0", "bnorm0"),
                edge("bnorm0", "activation0"),
                edge("activation0", "conv1"),
                edge("conv1", "bnorm1"),
                edge("bnorm1", "activation1"),
                edge("activation1", "output"),
            )
        )
    )
    pattern = (
        make_pattern()
        .add_nodes(
            (node("conv", "conv1d"), node("bnorm", "batchnorm1d"), node("act", "relu"))
        )
        .add_edges(
            (
                edge("start", "conv"),
                edge("conv", "bnorm"),
                edge("bnorm", "act"),
                edge("act", "end"),
            )
        )
    )

    def replacement(match):
        return (
            make_replacement()
            .add_node(node("fused_conv", "conv1d"))
            .add_node(node("activation", "relu"))
            .add_edge(edge("start", "fused_conv"))
            .add_edge(edge("fused_conv", "activation"))
            .add_edge(edge("activation", "end"))
        )

    spec = PatternRuleSpec(
        pattern=DummyPattern(
            graph=pattern,
            interface={"start", "end"},
            match=[
                {
                    "start": "input",
                    "conv": "conv0",
                    "bnorm": "bnorm0",
                    "act": "activation0",
                    "end": "conv1",
                },
            ],
        ),
        replacement_fn=replacement,
    )
    rule = PatternRule(spec)

    new_impl = rule.apply(impl)
    assert len(tuple(new_impl.nodes)) == len(tuple(impl.nodes)) - 1


def test_copy_edge_data_from_edges_connected_to_interface_nodes():
    impl = (
        make_network()
        .add_nodes(
            (
                node("conv0", "conv1d"),
                node("bnorm0", "batchnorm1d"),
                node("activation0", "relu"),
                node("conv1", "conv1d"),
                node("bnorm1", "batchnorm1d"),
                node("activation1", "relu"),
            )
        )
        .add_edges(
            (
                edge("input", "conv0", {"first": "first"}),
                edge("conv0", "bnorm0"),
                edge("bnorm0", "activation0"),
                edge("activation0", "conv1"),
                edge("conv1", "bnorm1", {"third": "third"}),
                edge("bnorm1", "activation1"),
                edge("activation1", "output"),
            )
        )
    )
    pattern = (
        make_pattern()
        .add_nodes(
            (node("conv", "conv1d"), node("bnorm", "batchnorm1d"), node("act", "relu"))
        )
        .add_edges(
            (
                edge("start", "conv"),
                edge("conv", "bnorm"),
                edge("bnorm", "act"),
                edge("act", "end"),
            )
        )
    )

    def replacement(match):
        return (
            make_replacement()
            .add_node(node("fused_conv", "conv1d"))
            .add_node(node("activation", "relu"))
            .add_edge(edge("start", "fused_conv", {"second": "second"}))
            .add_edge(edge("fused_conv", "activation"))
            .add_edge(edge("activation", "end"))
        )

    spec = PatternRuleSpec(
        pattern=DummyPattern(
            graph=pattern,
            interface={"start", "end"},
            match=[
                {
                    "start": "input",
                    "conv": "conv0",
                    "bnorm": "bnorm0",
                    "act": "activation0",
                    "end": "conv1",
                },
            ],
        ),
        replacement_fn=replacement,
    )
    rule = PatternRule(spec)

    new_impl = rule.apply(impl)
    assert {"second": "second", "third": "third"} == dict(
        new_impl.edges[("input", "fused_conv")].attributes
    ) | dict(new_impl.edges[("conv1", "bnorm1")].attributes)



from dataclasses import dataclass
from unittest import TestCase, main

from torch import Tensor, cat, tensor, testing, zeros
from torch.nn import BatchNorm1d, Linear, Sequential

from elasticai.creator.nn import Sequential as SequentialCreator
from elasticai.creator.nn.fixed_point import (
    BatchNormedLinear as BatchNormedLinearCreator,
)
from elasticai.creator.nn.fixed_point import (
    Linear as LinearCreator,
)


@dataclass
class SettingsTest:
    data_repeat_num: int = 3
    size_input_layer: int = 3
    size_output_layer: int = 1
    quant_width_total: int = 12
    quant_width_frac: int = 8


RecommendedSettings = SettingsTest()


def generate_test_data(settings: SettingsTest = RecommendedSettings) -> Tensor:
    outputs = zeros((1, settings.size_input_layer)).repeat(settings.data_repeat_num, 1)
    for idx in range(settings.data_repeat_num):
        outputs = cat((outputs, tensor([[-1.5, 0.5, 1.5]])), dim=0)
        outputs = cat((outputs, tensor([[1.5, 0.5, -1.5]])), dim=0)
    return outputs


def generate_torch_linear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = Linear(
        settings.size_input_layer, settings.size_output_layer, bias=True
    )
    linear_layer.bias.data = tensor(0.5)
    linear_layer.weight.data = tensor([[0.5, 1.0, -0.5]])
    return [linear_layer]


def generate_creator_linear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = LinearCreator(
        in_features=settings.size_input_layer,
        out_features=settings.size_output_layer,
        total_bits=settings.quant_width_total,
        frac_bits=settings.quant_width_frac,
        bias=True,
    )
    linear_layer.bias.data = tensor(0.5)
    linear_layer.weight.data = tensor([[0.5, 1.0, -0.5]])
    return [linear_layer]


def generate_torch_batchlinear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = generate_torch_linear(settings)

    batch1d_layer = BatchNorm1d(settings.size_output_layer)
    batch1d_layer.bias.data = tensor([0.25])
    batch1d_layer.weight.data = tensor([0.75])
    linear_layer.append(batch1d_layer)
    return linear_layer


def generate_creator_batchlinear(settings: SettingsTest = RecommendedSettings) -> list:
    batch1d_layer = BatchNormedLinearCreator(
        in_features=settings.size_input_layer,
        out_features=settings.size_output_layer,
        total_bits=settings.quant_width_total,
        frac_bits=settings.quant_width_frac,
        bias=True,
        bn_affine=True,
    )
    batch1d_layer.lin_bias.data = tensor([[0.5]])
    batch1d_layer.lin_weight.data = tensor([[0.5, 1.0, -0.5]])

    batch1d_layer.bn_bias.data = tensor([0.25])
    batch1d_layer.bn_weight.data = tensor([0.75])
    return [batch1d_layer]


def build_torch_sequential(nn_layer: list, do_training: bool = False) -> Sequential:
    model_build = Sequential()
    for layer in nn_layer:
        model_build.append(layer)

    if do_training:
        model_build.train()
    else:
        model_build.eval()
    return model_build


def build_creator_sequential(
    nn_layer: list, do_training: bool = False
) -> SequentialCreator:
    model_build = SequentialCreator()
    for layer in nn_layer:
        model_build.append(layer)

    if do_training:
        model_build.train()
    else:
        model_build.eval()
    return model_build


class TestCreatorBatchTraining(TestCase):
    model_torch = build_torch_sequential(
        generate_torch_batchlinear(), do_training=False
    )
    model_creator = build_creator_sequential(
        generate_creator_batchlinear(), do_training=False
    )
    data_in = generate_test_data()

    def test_data_input_size(self):
        testing.assert_close(self.data_in.size(1), RecommendedSettings.size_input_layer)

    def test_data_output_size_torch(self):
        testing.assert_close(
            self.model_torch(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_data_output_size_creator(self):
        testing.assert_close(
            self.model_creator(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_creator_batch(self):
        data_out_torch = self.model_torch(self.data_in)
        data_out_creator = self.model_creator(self.data_in)
        testing.assert_close(
            data_out_torch,
            data_out_creator,
            atol=1 / (2**RecommendedSettings.quant_width_frac),
            rtol=1 / (2**RecommendedSettings.quant_width_frac),
        )


class TestCreatorBatchEvaluation(TestCase):
    model_torch = build_torch_sequential(generate_torch_batchlinear(), do_training=True)
    model_creator = build_creator_sequential(
        generate_creator_batchlinear(), do_training=True
    )
    data_in = generate_test_data()

    def test_data_input_size(self):
        testing.assert_close(self.data_in.size(1), RecommendedSettings.size_input_layer)

    def test_data_output_size_torch(self):
        testing.assert_close(
            self.model_torch(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_data_output_size_creator(self):
        testing.assert_close(
            self.model_creator(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_creator_batch(self):
        data_out_torch = self.model_torch(self.data_in)
        data_out_creator = self.model_creator(self.data_in)
        testing.assert_close(
            data_out_torch,
            data_out_creator,
            atol=1 / (2**RecommendedSettings.quant_width_frac),
            rtol=1 / (2**RecommendedSettings.quant_width_frac),
        )


if __name__ == "__main__":
    main()



import math
from abc import abstractmethod
from collections import defaultdict
from typing import Protocol

from elasticai.creator.arithmetic import FxpConverter, FxpParams
from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.testing import Testbench
from elasticai.creator.vhdl.design.ports import Port


class Conv1dDesignProtocol(Protocol):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def input_signal_length(self) -> int: ...

    @property
    @abstractmethod
    def port(self) -> Port: ...

    @property
    @abstractmethod
    def kernel_size(self) -> int: ...

    @property
    @abstractmethod
    def in_channels(self) -> int: ...

    @property
    @abstractmethod
    def out_channels(self) -> int: ...


class Conv1dTestbench(Testbench):
    def __init__(self, name: str, uut: Conv1dDesignProtocol, fxp_params: FxpParams):
        self._converter = FxpConverter(
            FxpParams(
                total_bits=fxp_params.total_bits,
                frac_bits=fxp_params.frac_bits,
                signed=True,
            )
        )
        self._converter_for_batch = FxpConverter(
            FxpParams(total_bits=8, frac_bits=0, signed=True)
        )  # max for 255 lines of inputs
        self._name = name
        self._uut_name = uut.name
        self._input_signal_length = uut.input_signal_length
        self._in_channels = uut.in_channels
        self._out_channels = uut.out_channels
        self._x_address_width = uut.port["x_address"].width
        self._fxp_params = fxp_params
        self._kernel_size = uut.kernel_size
        self._output_signal_length = math.floor(
            self._input_signal_length - self._kernel_size + 1
        )
        self._y_address_width = uut.port["y_address"].width

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="testbench.tpl.vhd",
            parameters={
                "testbench_name": self.name,
                "input_signal_length": str(self._input_signal_length),
                "in_channels": str(self._in_channels),
                "out_channels": str(self._out_channels),
                "total_bits": str(self._fxp_params.total_bits),
                "x_address_width": str(self._x_address_width),
                "output_signal_length": str(self._output_signal_length),
                "y_address_width": str(self._y_address_width),
                "uut_name": self._uut_name,
            },
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)

    @property
    def name(self) -> str:
        return self._name

    def prepare_inputs(self, *inputs) -> list[dict]:
        batches = inputs[0]
        prepared_inputs = []
        for batch in batches:
            prepared_inputs.append({})
            for channel_id, channel in enumerate(batch):
                for time_step_id, time_step_val in enumerate(channel):
                    prepared_inputs[-1][f"x_{channel_id}_{time_step_id}"] = (
                        self._converter.rational_to_binary_string_vhdl(
                            time_step_val
                        ).replace('"', "")
                    )

        return prepared_inputs

    def parse_reported_content(self, content: list[str]) -> list[list[list[float]]]:
        """
        This function parses the reported content, which is just a list of strings.
        All lines starting with 'output_text:' are considered as a result of the testbench.
        These results will be stacked for each batch.
        So you get a list[list[list[float]]] which is similar to batch[out channels[output neurons[float]]].
        For each item reported it is checked if the string starts with 'result: '.
        If so the remaining part will be split by ','. The first number gives the batch. The second the result.
        The channels are greedy guessed.
        We do so by taking the first X-values, where X is the number of values per channel.
        After we have enough values for the channel, we increase the channel number.
        If you report not enough values per channel, this will look like the last channel has not reported enough values.
        """

        def split_list(a_list):
            out_channel_length = len(a_list) // self._out_channels
            new_list = list()
            out_channel_counter = (
                -1
            )  # start with -1 because it will be increased in first iteration of loop
            for i, value in enumerate(a_list):
                if i % out_channel_length == 0:
                    new_list.append(list())
                    out_channel_counter += 1
                new_list[out_channel_counter].append(value)
            return new_list

        results_dict = defaultdict(list)
        for line in map(str.strip, content):
            if line.startswith("result: "):
                batch_text = line.split(":")[1].split(",")[0][1:]
                output_text = line.split(":")[1].split(",")[1][0:]
                print("output_text: ", output_text)
                batch = int(self._converter_for_batch.binary_to_rational(batch_text))
                if "U" not in line.split(":")[1].split(",")[1][1:]:
                    output = self._converter.binary_to_rational(output_text)
                else:
                    output = output_text
                results_dict[batch].append(output)
        results = list()
        for x in results_dict.items():
            results.append(split_list(x[1]))
        if len(results) == 0:
            raise Exception(content)
        return list(results)



from .ir2verilog import (
    Code,
    Edge,
    Implementation,
    Ir2Verilog,
    Node,
    type_handler,
    type_handler_iterable,
)
from .templates import TemplateDirector, VerilogTemplate

__all__ = [
    "Ir2Verilog",
    "type_handler",
    "type_handler_iterable",
    "Node",
    "Edge",
    "Implementation",
    "Code",
    "TemplateDirector",
    "VerilogTemplate",
]



import re

import pytest

from elasticai.creator.ir2vhdl import Shape, VhdlNode

from ..vhdl_nodes.node_factory import (
    InstanceFactoryForCombinatorial,
)


@pytest.fixture(scope="class")
def node(raw_node):
    return InstanceFactoryForCombinatorial(raw_node)


def new_node(
    name: str,
    type: str,
    implementation: str,
    input_shape: Shape,
    output_shape: Shape,
    attributes: dict | None = None,
) -> VhdlNode:
    if attributes is None:
        attributes = {}
    n = VhdlNode(
        name=name, data=dict(type=type, implementation=implementation, **attributes)
    )
    n.input_shape = input_shape
    n.output_shape = output_shape

    return n


class TestStridingShiftRegister:
    @pytest.fixture(scope="class")
    def raw_node(self):
        conv0_channels = 2
        conv1_kernel_size = 3
        conv0_out_shape = Shape(conv0_channels, 1)
        conv1_in_shape = Shape(conv0_channels, conv1_kernel_size)
        n = new_node(
            name="a",
            type="striding_shift_register",
            implementation="striding_shift_register",
            input_shape=conv0_out_shape,
            output_shape=conv1_in_shape,
        )
        n.data["stride"] = 2
        return n

    def test_can_instantiate(self, node, raw_node):
        """
        data width is the size of a single time step, which
        translates to number of channels or depth of the preceding convolution

        num points is the number of time steps for the succeeding which corresponds to
        the kernel size of the succeeding convolution.
        """
        stride = raw_node.attributes["stride"]
        expected = (
            "a: entity work.striding_shift_register(rtl)",
            "generic map (",
            f"DATA_WIDTH => {raw_node.input_shape.depth},",
            f"NUM_POINTS => {raw_node.output_shape.width},",
            f"STRIDE => {stride}",
            ")",
            "port map (",
            "clk => clk,",
            "rst => rst,",
            "d_in => d_in_a,",
            "d_out => d_out_a,",
            "valid_in => valid_in_a,",
            "valid_out => valid_out_a",
            ");",
        )
        actual = tuple(line.strip() for line in node.instantiate())
        assert actual == expected

    def test_can_define_signals(self, node, raw_node):
        signals = tuple(line for line in node.define_signals())
        expected = (
            f"signal d_in_a : std_logic_vector({raw_node.input_shape.size()} - 1 downto 0) := (others => '0');",
            f"signal d_out_a : std_logic_vector({raw_node.output_shape.size()} - 1 downto 0) := (others => '0');",
            "signal valid_in_a : std_logic := '0';",
            "signal valid_out_a : std_logic := '0';",
        )
        assert signals == expected


class TestShiftRegister:
    @pytest.fixture(scope="class")
    def raw_node(self):
        conv0_channels = 2
        conv1_kernel_size = 4
        conv0_out_shape = Shape(conv0_channels, 2)
        conv1_in_shape = Shape(conv0_channels, conv1_kernel_size)
        n = new_node(
            name="a",
            type="shift_register",
            implementation="impl",
            input_shape=conv0_out_shape,
            output_shape=conv1_in_shape,
        )
        return n

    def test_can_instantiate(self, node, raw_node):
        expected = (
            "a: entity work.impl(rtl)",
            "generic map (",
            "DATA_WIDTH => 4,",
            "NUM_POINTS => 2",
            ")",
            "port map (",
            "clk => clk,",
            "rst => rst,",
            "d_in => d_in_a,",
            "d_out => d_out_a,",
            "valid_in => valid_in_a,",
            "valid_out => valid_out_a",
            ");",
        )
        actual = tuple(line.strip() for line in node.instantiate())
        assert actual == expected

    def test_can_define_signals(self, node):
        expected = (
            "signal d_in_a : std_logic_vector(4 - 1 downto 0) := (others => '0');",
            "signal d_out_a : std_logic_vector(8 - 1 downto 0) := (others => '0');",
            "signal valid_in_a : std_logic := '0';",
            "signal valid_out_a : std_logic := '0';",
        )
        signals = tuple(line for line in node.define_signals())

        assert signals == expected


class TestSlidingWindow:
    @pytest.fixture(scope="class")
    def raw_node(self):
        return new_node(
            name="a",
            type="sliding_window",
            input_shape=Shape(4, 2),
            output_shape=Shape(4, 1),
            implementation="impl",
            attributes={"stride": 2},
        )

    def test_can_instantiate(self, node):
        expected = (
            "a: entity work.impl(rtl)",
            "generic map (",
            "STRIDE => 8,",
            "INPUT_WIDTH => 8,",
            "OUTPUT_WIDTH => 4",
            ")",
            "port map (",
            "clk => clk,",
            "rst => rst,",
            "d_in => d_in_a,",
            "d_out => d_out_a,",
            "valid_in => valid_in_a,",
            "valid_out => valid_out_a",
            ");",
        )
        actual = tuple(line.strip() for line in node.instantiate())
        assert actual == expected

    def test_raise_error_for_incompatible_shapes(self):
        node_with_incompatible_shapes = new_node(
            name="a",
            type="sliding_window",
            input_shape=Shape(5, 2),
            output_shape=Shape(3, 2),
            implementation="impl",
            attributes={"stride": 2},
        )
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Found incompatible input output shapes for sliding_window. Total input size has to be an integer multiple of total output size, but found output=Shape(depth=3, width=2) and input=Shape(depth=5, width=2)."
            ),
        ):
            InstanceFactoryForCombinatorial(node_with_incompatible_shapes).instantiate()

    def test_warn_about_technically_compatible_but_semantically_wrong_shapes(self):
        node_with_incompatible_shapes = new_node(
            name="a",
            type="sliding_window",
            input_shape=Shape(2, 4),
            output_shape=Shape(1, 2),
            implementation="impl",
            attributes={"stride": 2},
        )
        with pytest.warns(
            match=re.escape(
                'Detected mismatching input output shapes for sliding_window for node "a". Depth of output and input shape should usually be equal, but found output=Shape(depth=1, width=2) and input=Shape(depth=2, width=4).'
            ),
        ):
            InstanceFactoryForCombinatorial(node_with_incompatible_shapes).instantiate()



from os.path import exists
from pathlib import Path

import pytest

from elasticai.creator.file_generation.resource_utils import (
    get_full_path,
)
from elasticai.creator.testing.cocotb_runner import (
    run_cocotb_sim,
    run_cocotb_sim_for_src_dir,
)

_path2src = str(
    Path(get_full_path("tests.unit_tests.testing", "cocotb_runner_tb.py")).parent
)
_path2tb = "tests.unit_tests.testing.cocotb_runner_tb"


@pytest.mark.simulation
def test_verilog_without_defines_without_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        defines={},
        params={},
    )


@pytest.mark.simulation
def test_verilog_without_defines_with_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={"BITWIDTH": 4, "SCALE": 2},
        defines={},
    )


@pytest.mark.simulation
def test_verilog_with_defines_without_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={"DEFINE_TEST": True},
    )


@pytest.mark.simulation
def test_verilog_with_defines_with_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={"BITWIDTH": 4, "SCALE": 2},
        defines={"DEFINE_TEST": True},
    )


@pytest.mark.simulation
def test_verilog_with_waveforms_default() -> None:
    path = run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={},
    )
    assert exists(path / "waveforms.vcd")


@pytest.mark.simulation
def test_verilog_with_waveforms_external() -> None:
    path = run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.v"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={},
        waveform_save_dst="build_test",
    )
    assert exists(path / "waveforms.vcd")


@pytest.mark.simulation
def test_vhdl_without_defines_without_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={},
    )


@pytest.mark.simulation
def test_vhdl_without_defines_with_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={"BITWIDTH": 4, "SCALE": 2},
        defines={},
    )


@pytest.mark.simulation
def test_vhdl_with_defines_without_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={"DEFINE_TEST": True},
    )


@pytest.mark.simulation
def test_vhdl_with_defines_with_params() -> None:
    run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={"BITWIDTH": 4, "SCALE": 2},
        defines={"DEFINE_TEST": True},
    )


@pytest.mark.simulation
def test_verilog_with_run_direct() -> None:
    run_cocotb_sim(
        src_files=[f"{_path2src}/cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        params={},
        defines={},
    )


@pytest.mark.simulation
def test_vhdl_with_run_direct() -> None:
    run_cocotb_sim(
        src_files=[f"{_path2src}/cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        params={},
        defines={},
    )


@pytest.mark.simulation
def test_vhdl_with_waveforms_normal() -> None:
    path = run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={},
        waveform_save_dst="",
    )
    assert exists(path / "waveforms.vcd")


@pytest.mark.simulation
def test_vhdl_with_waveforms_external() -> None:
    path = run_cocotb_sim_for_src_dir(
        src_files=["cocotb_runner_tb.vhd"],
        top_module_name="COCOTB_TEST",
        cocotb_test_module=_path2tb,
        path2src=_path2src,
        params={},
        defines={},
        waveform_save_dst="build_test",
    )
    assert exists(path / "waveforms.vcd")



from unittest import TestCase

from elasticai.creator_plugins.grouped_filter.src.iterator_utils import batched
from elasticai.creator_plugins.grouped_filter.src.tensor import (
    reshape_flat_CxN_groupwise,
    structure_flat_CxN,
)


class ReshapeCxNTest(TestCase):
    def test_interpret_flat_vector_as_2x6(self):
        expected = ((0, 1, 2, 3, 4, 5), (6, 7, 8, 9, 10, 11))
        self.assertEqual(expected, structure_flat_CxN(tuple(range(12)), channels=6))

    def test_interpret_flat_vector_as_3x4(self):
        expected = (
            (
                0,
                1,
                2,
            ),
            (3, 4, 5),
            (6, 7, 8),
            (9, 10, 11),
        )
        self.assertEqual(expected, structure_flat_CxN(tuple(range(12)), channels=3))


class BatchedTest(TestCase):
    def test_batched(self):
        expected = ([0, 1], [2, 3], [4, 5])
        actual = tuple(batched(list(range(6)), 2))
        self.assertEqual(expected, actual)


class GroupWiseConnectionTest(TestCase):
    def test_interpret_flat_vector_as_6x2_and_reshape_to_3_groups(self):
        expected = ((0, 1, 6, 7), (2, 3, 8, 9), (4, 5, 10, 11))

        self.assertEqual(
            expected, reshape_flat_CxN_groupwise(tuple(range(12)), channels=6, groups=3)
        )

    def test_interpret_flat_vector_as_4x3_and_reshape_to_4_groups(self):
        expected = ((0, 4, 8), (1, 5, 9), (2, 6, 10), (3, 7, 11))
        self.assertEqual(
            expected, reshape_flat_CxN_groupwise(tuple(range(12)), channels=4, groups=4)
        )

    def test_interpret_flat_vector_as_6x4_and_reshape_to_2_groups(self):
        expected = (
            (0, 1, 2, 6, 7, 8, 12, 13, 14, 18, 19, 20),
            (3, 4, 5, 9, 10, 11, 15, 16, 17, 21, 22, 23),
        )
        self.assertEqual(
            expected, reshape_flat_CxN_groupwise(tuple(range(24)), channels=6, groups=2)
        )
